@@ -1,5 +1,5 @@
-/*
- * $Id: TestSuite.java,v 1.13 1999/12/19 17:43:31 metlov Exp $
+/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
+ * $Id: TestSuite.java,v 1.15 2000/02/12 17:31:32 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
@@ -25,7 +25,7 @@ public class TestSuite {
   static int successes=0;
   static int failures=0;
   static boolean abort_at_error=false;// Set if You want to abort at a failing
-                                      // test, works only for testExpression().
+  // test, works only for testExpression().
   
   public static void main(String[] args) {
 
@@ -42,7 +42,7 @@ public class TestSuite {
     };
     Library lib=new Library(staticLib,null);
     try {
-    lib.markStateDependent("random",null);
+      lib.markStateDependent("random",null);
     } catch (NoSuchMethodException e) {
       // Can't be also
     };
@@ -73,11 +73,11 @@ public class TestSuite {
     testError("0.5+(floatp)0.4D",null,lib,6,o);
     testError("0.5+(boolean)0.4D",null,lib,6,o);
     testError("23?1:2",null,lib,3,o);
-    testError("true?\" \":' '",null,lib,10,o);
-    testError("true?\" \":makeDoubleObject(1.0)",null,lib,30,o);
+    testError("true?\" \":' '",null,lib,5,o);
+    testError("true?\" \":makeDoubleObject(1.0)",null,lib,5,o);
     
     o.println(); o.println("----- TESTING STATIC OPTIMIZATIONS ------");
-    o.println("lines starting with <number>| denote optimization passes.");
+    o.println("lines starting with <number>| denote constants folding passes.");
     o.println("Expression on each line is evaluated 20 times to provoke some JIT compilers.");
     o.println();
     
@@ -87,45 +87,45 @@ public class TestSuite {
     testExpression("2.0F*2.0F",new Float(4.0F),null,null,lib,o);
     testExpression("sin(1)",new Double(Math.sin(1.0)),null,null,lib,o);
     testExpression("pow(sin(1),2)+pow(cos(1),2)",new Double(1.0),null,null,
-		   lib,o);
+                   lib,o);
     testExpression("min(1+2*2,(1+2)*2)",new Integer(5),null,null,lib,o);
     testExpression("7+4-6",new Double(5.0),Double.TYPE,null,lib,o);
     testExpression("true&&false||false&&true||false",Boolean.FALSE,
-		   null,null,lib,o);
+                   null,null,lib,o);
 
     testExpression("(1<<2L==4)&&(-1>>5==-1)&&(-1>>>1==0x7FFFFFFF)",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("(-1>>>1==0x5FFFFFFF)",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("(-1L>>>1L==0x7FFFFFFFFFFFFFFFL)",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("1+2>2==true",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("true?false:true?true:false",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("false?true:true?false:true",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     
     testExpression("(1==1)&&(max(~bool2int(1<=2),~bool2int(2>=3))!=-1)",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("(-1==-1)&&(max(~(1<=2?1:0),~(2>=3?1:0))!=-1)",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("!((!true)&&(!true))",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("!(!(false&&false&&false)&&!false)",
-		   Boolean.FALSE,null,null,lib,o);
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("(!(5+1>5)?1:2)==2",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testFullLogic("!(!(_a&&_b&&_c)&&!_d)",4,lib,o,false);
     
@@ -150,34 +150,34 @@ public class TestSuite {
     testFullLogic("_a?_b:_c?_d:_e",5,lib,o,false);
 
     testExpression("\"aa\"+\"bb\"+\"cc\"",
-		   "aabbcc",null,null,lib,o);
+                   "aabbcc",null,null,lib,o);
 
     testExpression("\"str\"+true+1+20L+6.0F+7.0D",
-		   "strtrue1206.07.0",null,null,lib,o);
+                   "strtrue1206.07.0",null,null,lib,o);
 
     testExpression("\"str\"+(2+3)","str5",null,null,lib,o);
 
     testExpression("((~(~((((1+2-2)*2/2)^55)^55)))|1&2)==1",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("((~(~((((1L+2L-2L)*2L/2L)^55L)^55L)))|1L&2L)==1L",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("((10/3)*3+10%3)==10",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("((10L/3)*3+10%3L)==10L",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("((1>5?0:10F)/3)*3+10F%3",
-		   new Float(11.0F),null,null,lib,o);
+                   new Float(11.0F),null,null,lib,o);
 
     testExpression("((1>5?0:10D)/3)*3+10D%3",
-		   new Double(11.0),null,null,lib,o);
+                   new Double(11.0),null,null,lib,o);
 
     testExpression("round(((10-5)==5?(6.0-((8==8)?5.0:1))*2.0:1.0+"+
-		   "2.0-1.0)/2.0)==1",
-		   Boolean.TRUE,null,null,lib,o);
+                   "2.0-1.0)/2.0)==1",
+                   Boolean.TRUE,null,null,lib,o);
 
 
     // Tests if NaN (floating point "not a number") handling conforms to JLS
@@ -212,48 +212,48 @@ public class TestSuite {
     // to remove the above since ... well... it makes testsuite runs
     // more impressive ;)))
     testExpression("(NaNd>1)||(NaNd<1)||(NaNd>=1)||(NaNd<=1)||"+
-		   "(NaNd==1)||!(NaNd!=1)",
-		   Boolean.FALSE,null,null,lib,o);
+                   "(NaNd==1)||!(NaNd!=1)",
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("(1>NaNd)||(1<NaNd)||(1>=NaNd)||(1<=NaNd)||"+
-		   "(1==NaNd)||!(1!=NaNd)",
-		   Boolean.FALSE,null,null,lib,o);
+                   "(1==NaNd)||!(1!=NaNd)",
+                   Boolean.FALSE,null,null,lib,o);
 
 
     testExpression("(NaNf>1)||(NaNf<1)||(NaNf>=1)||(NaNf<=1)||"+
-		   "(NaNf==1)||!(NaNf!=1)",
-		   Boolean.FALSE,null,null,lib,o);
+                   "(NaNf==1)||!(NaNf!=1)",
+                   Boolean.FALSE,null,null,lib,o);
 
     testExpression("(1>NaNf)||(1<NaNf)||(1>=NaNf)||(1<=NaNf)||"+
-		   "(1==NaNf)||!(1!=NaNf)",
-		   Boolean.FALSE,null,null,lib,o);
+                   "(1==NaNf)||!(1!=NaNf)",
+                   Boolean.FALSE,null,null,lib,o);
     
     testExpression("indexOf(\"aaaca\",'c')+1==4 && "+
-		   "indexOf(1>2?\"cc\":\"aaaca\",'c')+1==4",
-		   Boolean.TRUE,null,null,lib,o);
+                   "indexOf(1>2?\"cc\":\"aaaca\",'c')+1==4",
+                   Boolean.TRUE,null,null,lib,o);
 
     testExpression("round((float)4)",
-		   new Integer(4),null,null,lib,o);
+                   new Integer(4),null,null,lib,o);
 
     testExpression("(int)4.0",
-		   new Integer(4),null,null,lib,o);
+                   new Integer(4),null,null,lib,o);
 
     
     testExpression("7+(int)4-(int)6.0+1-(int)round((double)((float)((long)1+0)+0)+0)",
-		   new Integer(5),null,null,lib,o);
+                   new Integer(5),null,null,lib,o);
     
     testExpression("~0",new Integer(-1),null,null,lib,o);
 		   
     testExpression("~0",new Integer(-1),null,null,lib,o);
 
     testExpression("substring(\"abbb\"+\"ccc\"+'d'+(1>2?\"e\":\"d\"),1)",
-		   "bbbcccdd",null,null,lib,o);
+                   "bbbcccdd",null,null,lib,o);
 
     testExpression("equals(substring(\"abbb\",1),\"bbb\")||false",
-		   Boolean.TRUE,null,null,lib,o);
+                   Boolean.TRUE,null,null,lib,o);
 
     o.println(); 
-    o.println("----- TESTING DYNAMIC OPTIMIZATIONS (x=5.0D) ------");
+    o.println("----- TESTING VIRTUAL METHODS (x=5.0D) ------");
     o.println();
 
     // Construct a new library
@@ -277,7 +277,7 @@ public class TestSuite {
     testExpression("-x+(255+5+7+9)",new Double(255+7+9),null,rtp,lib,o);
     testExpression("5*x-66",new Double(25-66),Double.TYPE,rtp,lib,o);
     testExpression("7+(int)4-(int)6.0+(int)x-(int)round((double)((float)((long)x+1)+2)+3)+6",
-		   new Integer(5),null,rtp,lib,o);
+                   new Integer(5),null,rtp,lib,o);
 
     o.println(); o.println("----- VARIABLE AND ARRAY ACCESS ------");
     o.println();
@@ -293,8 +293,8 @@ public class TestSuite {
     testExpression("throw_arg_eq_4(6-2)",null,null,rtp,lib,o);
     
     o.println(); o.println("^^^^^^^^^^^^ SCORE success/failure = "+
-			   successes+"/"+failures+
-			   " ^^^^^^^^^^^^" );
+                           successes+"/"+failures+
+                           " ^^^^^^^^^^^^" );
 
   };
 
@@ -311,7 +311,7 @@ public class TestSuite {
   // This function does not analyze syntax of expression so be sure not to have
   // underscores ("_") in the names of functions.
   private static void testFullLogic(String expr,int bits,Library lib, 
-				    PrintStream o,boolean showcases) {
+                                    PrintStream o,boolean showcases) {
     int cases=1<<bits;
     o.print("*** : FULL LOGIC TEST \""); o.print(expr); 
     o.println("\" . ( "+cases+" cases ).");    
@@ -322,96 +322,91 @@ public class TestSuite {
       for(int i=0;i<bits;i++) vars[i]=((ccase>>>i & 0x00000001)>0?true:false);
       StringBuffer cexpr=new StringBuffer();
       for (int i=0;i<expr.length();i++) {
-	char currchar=expr.charAt(i);
-	if (currchar=='_') {
-	  currchar=expr.charAt(++i);
-	  int varnum=currchar-'a';
-	  if (vars[varnum]) cexpr.append("true "); else  cexpr.append("false");
-	} else cexpr.append(currchar);
+        char currchar=expr.charAt(i);
+        if (currchar=='_') {
+          currchar=expr.charAt(++i);
+          int varnum=currchar-'a';
+          if (vars[varnum]) cexpr.append("true "); else  cexpr.append("false");
+        } else cexpr.append(currchar);
       };
 
       // Now we need to calculate cexpr
       
       // First parse it
-      OpenOptimizer op=null;
+      OPlist op=null;
       try {
-	op=fillOptimizer(cexpr.toString(),lib,null);
+        op=OpenEvaluator.parse(cexpr.toString(),lib,null);
       } catch (CompilationException ce) {
-	o.print("--- COMPILATION ERROR :");
-	o.println(ce.getMessage());
-	o.print("                       ");
-	o.println(cexpr.toString());
-	int column=ce.getColumn(); // Column, where error was found
-	for(int i=0;i<column+23-1;i++) System.err.print(' ');
-	o.println('^');
-	o.println("Unexpected syntax error on supposingly correct"+
-		  " expression.");
-	FAIL(o);
-	return;
+        o.print("--- COMPILATION ERROR :");
+        o.println(ce.getMessage());
+        o.print("                       ");
+        o.println(cexpr.toString());
+        int column=ce.getColumn(); // Column, where error was found
+        for(int i=0;i<column+23-1;i++) System.err.print(' ');
+        o.println('^');
+        o.println("Unexpected syntax error on supposingly correct"+
+                  " expression.");
+        FAIL(o);
+        return;
       };
       
       // Make optimization iterations
       Object result=null;      
-      boolean optimized=true;
-      int iterations=0;
-      while(optimized) {
-	Object result1=null;      
-	try {
-	  CompiledExpression expr_c=op.compile();
 
-	  // Execute several times to enable JIT compilation.
-	  // Some JITs compile methods if they are run more than once
-	  for(int acounter=0;acounter<20;acounter++) {
-	    result1=expr_c.evaluate(null);
-	  };
-	} catch (Throwable e) {
-	  o.println(cexpr.toString());
-	  o.println("Exception emerged during compilation/evaluation.");
-	  o.print("      ");e.printStackTrace();
-	  FAIL(o);
-	  return;
-	};
+      for(int iteration=0;iteration<2;iteration++) {
+        Object result1=null;      
+        try {
+          CompiledExpression expr_c=OpenEvaluator.compile(op);
+
+          // Execute several times to enable JIT compilation.
+          // Some JITs compile methods if they are run more than once
+          for(int acounter=0;acounter<20;acounter++) {
+            result1=expr_c.evaluate(null);
+          };
+        } catch (Throwable e) {
+          o.println(cexpr.toString());
+          o.println("Exception emerged during compilation/evaluation.");
+          o.print("      ");e.printStackTrace();
+          FAIL(o);
+          return;
+        };
 	
-	if (result!=null) {
-	  if (!result.equals(result1)) {
-	    o.println(cexpr.toString());
-	    o.println("Expression gave inconsistent result "+result1+
-		      " ( previous result was "+result+" ).");
-	    o.println(op.toString());
-	    FAIL(o);
-	    return;
-	  };
-	};
-	result=result1;
+        if (result!=null) {
+          if (!result.equals(result1)) {
+            o.println(cexpr.toString());
+            o.println("Expression gave inconsistent result "+result1+
+                      " ( previous result was "+result+" ).");
+            o.println(op.toString());
+            FAIL(o);
+            return;
+          };
+        };
+        result=result1;
 
-	optimized=op.optimizeIteration();
-	iterations++;
+        if (iteration==0) op.performCF();
       };
       
-      if (iterations<=1) 
-	o.println("WARNING : only one optimizing iteration was performed.");
-
       if (showcases) {
-	o.print(cexpr.toString()); 
-	o.print(" == "); 
-	o.println(result.toString());
+        o.print(cexpr.toString()); 
+        o.print(" == "); 
+        o.println(result.toString());
       };
     };
     OK(o);
   };
 
   private static void testExpression(String expr, Object tobe, Class fixType,
-				     Object[] runtimeParameters,
-				     Library lib,   PrintStream o ) {
+                                     Object[] runtimeParameters,
+                                     Library lib,   PrintStream o ) {
     o.print("*** : \""); o.print(expr); 
     if (tobe != null) {
       o.print("\" = "); o.println(tobe);
     } else {
       o.println("\"   Should throw an exception at run time.");
     };
-    OpenOptimizer op=null;
+    OPlist op=null;
     try {
-      op=fillOptimizer(expr,lib,fixType);
+      op=OpenEvaluator.parse(expr,lib,fixType);
     } catch (CompilationException ce) {
       o.print("--- COMPILATION ERROR :");
       o.println(ce.getMessage());
@@ -427,50 +422,46 @@ public class TestSuite {
     
     boolean testok=true;
     
-    boolean optimized=true;
-    int iteration=0;
-    while(optimized) {
+    for(int iteration=0;iteration<2;iteration++) {
       String message=""+iteration+" |"+op.toString();
       o.print(message);
       for (int k=message.length();k<59;k++) o.print(' ');
 
       Object result=null;
       try {
-	CompiledExpression expr_c=op.compile();
-	
-	// Execute several times to enable JIT compilation.
-	// Some JITs compile methods if they are run more than once
-	for(int acounter=0;acounter<20;acounter++) {
-	  result=expr_c.evaluate(runtimeParameters);
-	};
+        CompiledExpression expr_c=OpenEvaluator.compile(op);
 	
+        // Execute several times to enable JIT compilation.
+        // Some JITs compile methods if they are run more than once
+        for(int acounter=0;acounter<20;acounter++) {
+          result=expr_c.evaluate(runtimeParameters);
+        };
       } catch (Throwable e) {
-	if (tobe==null) {
-	  o.println("EXPECTED EXCEPTION.");
-	  o.print("      ");o.println(e.getMessage());
-	} else {
-	  o.println("Exception emerged during compilation/evaluation.");
-	  o.print("      ");e.printStackTrace();
-	};
-	testok=(tobe==null);
+        if (tobe==null) {
+          o.println("EXPECTED EXCEPTION.");
+          o.print("      ");o.println(e.getMessage());
+        } else {
+          o.println("Exception emerged during compilation/evaluation.");
+          o.print("      ");e.printStackTrace();
+        };
+        testok=(tobe==null);
       };
       
       if (tobe!=null) {
-	if (result!=null) {
-	  o.print(" ="); o.println(result);
-	  testok=testok && (result.equals(tobe));
-	  if (!result.equals(tobe) && (abort_at_error)) System.exit(1);
-	} else {
-	  o.println("NO RESULT");
-	  testok=false;
-	};
+        if (result!=null) {
+          o.print(" ="); o.println(result);
+          testok=testok && (result.equals(tobe));
+          if (!result.equals(tobe) && (abort_at_error)) System.exit(1);
+        } else {
+          o.println("NO RESULT");
+          testok=false;
+        };
       } else {
-	testok=(result==null);
-	if (result!=null) o.println(" ="+result.toString());
+        testok=(result==null);
+        if (result!=null) o.println(" ="+result.toString());
       };
       
-      optimized=op.optimizeIteration();
-      iteration++;
+      if (iteration==0) op.performCF();
     };
     
     if (testok) OK(o); else FAIL(o);
@@ -478,12 +469,12 @@ public class TestSuite {
   };
   
   private static void testError(String expr,Class fixType, Library lib, 
-				int errcol, PrintStream o) {
+                                int errcol, PrintStream o) {
     o.print("*** : \""); o.print(expr); o.println('"');
 
     CompilationException ce=null;
     try {
-      OpenOptimizer op=fillOptimizer(expr,lib,fixType);
+      OPlist op=OpenEvaluator.parse(expr,lib,fixType);
     } catch (CompilationException e) {
       ce=e;
     };
@@ -499,9 +490,9 @@ public class TestSuite {
       o.print("MESSAGE: "); o.println(ce.getMessage());
       
       if (ce.getColumn()!=errcol) {
-	o.print("Error detected at column "); o.print(ce.getColumn());
-	o.print(" while it should be at "); o.print(errcol); o.println(" .");
-	FAIL(o);
+        o.print("Error detected at column "); o.print(ce.getColumn());
+        o.print(" while it should be at "); o.print(errcol); o.println(" .");
+        FAIL(o);
       } else OK(o);
 
     };
@@ -518,35 +509,22 @@ public class TestSuite {
     o.println(" TEST PASSED."); //o.println();
   };
   
-
-
-  // Mostly copied from jel.Evaluator
-  static OpenOptimizer fillOptimizer(String expression, Library lib, Class rt) 
-       throws CompilationException {
-	 EC ec=new EC(new java.io.StringReader(expression));
-    OpenOptimizer o=new OpenOptimizer(lib);
-    try {
-      ec.expression(o,rt);
-    } catch (ParseException pe) {
-      throw new CompilationException(pe);
-    } catch (TokenMgrError tme) {
-      throw new CompilationException(tme);
-    };
-    return o;
-  };
-
 };
 
-class OpenOptimizer extends Optimizer {
+class OpenEvaluator extends Evaluator {
   
-  OpenOptimizer(Library lib) {super(lib);};
-  
-  public boolean optimizeIteration() {
-    try {
-      return super.optimizeIteration(super.code);
-    } catch (Throwable e) {
-      return false;
-    };
+  public static OPlist parse(String expression, Library lib,
+                             Class resultType) throws CompilationException {
+	return Evaluator.parse(expression,lib,resultType);
+  };
+
+  public static CompiledExpression compile(OPlist list) {
+	byte[] image=Evaluator.getImage(list);
+	try {
+      return (CompiledExpression)(ImageLoader.load(image)).newInstance();
+	} catch (Exception exc) {
+      return null;
+	};
   };
   
 };
