@@ -1,11 +1,11 @@
-/*
- * $Id: Evaluator.java,v 1.6 1999/12/19 17:43:31 metlov Exp $
+/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
+ * $Id: Evaluator.java,v 1.7 2000/02/11 22:08:18 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
  *    http://galaxy.fzu.cz/JEL/
  *
- * (c) 1998,1999 by Konstantin Metlov(metlov@fzu.cz);
+ * (c) 1998 -- 2000 by Konstantin Metlov(metlov@fzu.cz);
  *
  * JEL is Distributed under the terms of GNU General Public License.
  *    This code comes with ABSOLUTELY NO WARRANTY.
@@ -15,9 +15,8 @@
 package gnu.jel; 
 
 import gnu.jel.generated.EC;
-import gnu.jel.generated.ParseException;
-import gnu.jel.generated.TokenMgrError;
 import gnu.jel.debug.Debug;
+import gnu.jel.reflect.*;
 
 /**
  * This is the main front end to JEL.
@@ -52,6 +51,57 @@ import gnu.jel.debug.Debug;
  * @see gnu.jel.CompiledExpression
  */
 public class Evaluator {
+
+  protected static ClassFile cf_orig;
+  protected static int retID_patchback=0;
+  protected static LocalMethod[] eval_methods=
+    new LocalMethod[TypesStack.primitiveTypes.length];
+
+  static {
+    try {
+      // prepare eval methods
+      Class[] paramsE=new Class[1];
+      paramsE[0]=(new Object[0]).getClass();
+      for(int i=0;i<TypesStack.primitiveTypes.length-1;i++) {
+        String name="evaluate";
+        Class cls=TypesStack.primitiveTypes[i];
+        if (i!=8) 
+          name=name+'_'+TypesStack.primitiveTypeNames[i];
+        else 
+          cls=(new Object()).getClass();
+        eval_methods[i]=new LocalMethod(0x0001,cls,name,paramsE,null);
+      };
+      
+      Class cmplExpr=Class.forName("gnu.jel.CompiledExpression");
+      ClassFile cf=new ClassFile(0x0001,"dump",cmplExpr,null,null);
+      // public 
+      LocalMethod cnstr=
+        new LocalMethod(0x0001,Void.TYPE,"<init>",null,null);
+      cf.newMethod(cnstr,null);
+      cf.code(0x2a);                //| aload_0  ;loads "this"
+      cf.typesStk.push(null);
+      Method supInit=new Method(cmplExpr.getConstructor(new Class[0]));
+      cf.code(0xb7);                //| invokespecial
+      cf.codeI(cf.getIndex(supInit,10));  //|    super();
+      cf.typesStk.pop();
+      cf.code(0xb1);                //| return void
+      
+      LocalMethod getType=
+      new LocalMethod(0x0001,Integer.TYPE,"getType",null,null);
+      cf.newMethod(getType,null);
+      cf.code(0x10);                //| bipush
+      retID_patchback=cf.textData.size();
+      cf.code(8);                   //    type placeholder
+      cf.typesStk.push(Integer.TYPE);
+      
+      cf.code(0xAC);                //| ireturn
+      cf.typesStk.pop();
+      
+      cf_orig=(ClassFile)cf.clone();
+    } catch (Exception exc) {
+      if (Debug.enabled) Debug.reportThrowable(exc);
+    };
+  };
   
   /**
    * Compiles expression, resolving the function names in the library.
@@ -66,11 +116,18 @@ public class Evaluator {
    * @see gnu.jel.CompiledExpression#evaluate
    */
   public static CompiledExpression compile(String expression, Library lib,
-					   Class resultType) 
-       throws CompilationException {
-    return compileBits(expression,lib,resultType).getExpression();
+                                           Class resultType) 
+    throws CompilationException {
+    byte[] image=compileBits(expression,lib,resultType);
+    try {
+      return (CompiledExpression)(ImageLoader.load(image)).newInstance();
+    } catch (Exception exc) {
+      if (Debug.enabled)
+        Debug.reportThrowable(exc);
+      return null;
+    };
   };
-
+  
   /**
    * Compiles expression, resolving the function names in the library.
    * <P>This variant of compile allows to store expressions in a 
@@ -85,12 +142,10 @@ public class Evaluator {
    *  syntactically or semantically correct.
    * @see gnu.jel.CompiledExpression#evaluate
    */
-  public static ExpressionBits compileBits(String expression, Library lib,
-					   Class resultType) 
-       throws CompilationException {
-    Optimizer o=fillOptimizer(expression, lib,resultType);
-    o.optimize(1);
-    return o.compileBits();
+  public static byte[] compileBits(String expression, Library lib,
+                                   Class resultType) 
+    throws CompilationException {
+    return getImage(parse(expression,lib,resultType));
   };
 
   /**
@@ -104,7 +159,7 @@ public class Evaluator {
    * @see gnu.jel.CompiledExpression#evaluate
    */
   public static CompiledExpression compile(String expression, Library lib)
-       throws CompilationException {
+    throws CompilationException {
     return compile(expression, lib, null);
   };
 
@@ -120,28 +175,32 @@ public class Evaluator {
    *  syntactically or semantically correct.
    * @see gnu.jel.CompiledExpression#evaluate
    */
-  public static ExpressionBits compileBits(String expression, Library lib)
-       throws CompilationException {
+  public static byte[] compileBits(String expression, Library lib)
+    throws CompilationException {
     return compileBits(expression, lib, null);
   };
-  
 
-  // This is for convenient use in the internal optimizer tests
-  static Optimizer fillOptimizer(String expression, Library lib, 
-				 Class resultType) 
-       throws CompilationException {
-    EC ec=new EC(new java.io.StringReader(expression));
-    Optimizer o=new Optimizer(lib);
-    try {
-      ec.expression(o,resultType);
-    } catch (ParseException pe) {
-      throw new CompilationException(pe);
-    } catch (TokenMgrError tme) {
-      throw new CompilationException(tme);
-    };
-    return o;
+  protected static OPlist parse(String expression, Library lib,
+                                Class resultType) throws CompilationException {
+    return 
+      (new EC(new java.io.StringReader(expression))).parse(resultType,lib);
   };
 
+  protected static byte[] getImage(OPlist list) {
+    int retID=((OPunary)list.getLast()).resID;
+
+    ClassFile cf=(ClassFile)cf_orig.clone();
+
+    // set return type
+    cf.textData.patchAddress(retID_patchback,(byte)retID);
+
+    // add the evaluate method
+    cf.newMethod(eval_methods[retID],null);
+    list.compile(cf);
+
+    return cf.getImage();
+  };
+  
 };
 
 
