@@ -1,11 +1,11 @@
-/*
- * $Id: Library.java,v 1.2 1999/12/19 17:43:31 metlov Exp $
+/*-*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
+ * $Id: Library.java,v 1.4 2000/02/11 22:08:18 metlov Exp $
  *
  * This file is part of the Java Expressions Library (JEL).
  *   For more information about JEL visit :
  *    http://galaxy.fzu.cz/JEL/
  *
- * (c) 1998,1999 by Konstantin Metlov(metlov@fzu.cz);
+ * (c) 1998 -- 2000 by Konstantin Metlov(metlov@fzu.cz);
  *
  * JEL is Distributed under the terms of GNU General Public License.
  *    This code comes with ABSOLUTELY NO WARRANTY.
@@ -16,47 +16,41 @@ package gnu.jel;
 
 import gnu.jel.debug.Debug;
 import gnu.jel.debug.Tester;
+import gnu.jel.reflect.*;
 import java.util.Hashtable;
 import java.util.Vector;
 import java.util.Enumeration;
-import java.lang.reflect.Method;
-import java.lang.reflect.Field;
-import java.lang.reflect.Modifier;
 
 /**
- * A Library of the methods, callable from compiled expressions.
- * <P> There are following constraints on the functions in the library
- * classes:
- * <UL> 
- * <LI>Functions can not return objects which are changeable by other
- * functions in the library. The best is if the functions do not return
- * mutable objects at all.
- * <LI>If a static function has the internal state it should be explicitly
- * marked as such (see markStateDependent(String name,Class[] params)). 
- * Evaluation of stateless functions of the constant arguments will be
- * performed at compile time. All static library functions are initially
- * assumed to be state independent. The example of the function which will
- * certainly have to be made state dependent is <TT>java.lang.Math.random</TT>
- * </UL>
+ * A namespace for JEL expressions.
+ * <P> There are two types of members in the library, those which are stateless
+ * (i.e. their value depends only on their arguments, if there are any) and
+ * stateful (also called here dynamic). The result of evaluation of 
+ * stateful members may depend on other factors besides their arguments.
+ *
+ * <P>Examples of possible stateless members are : <TT>Math.sin(double)</TT>,
+ * <TT>Math.PI</TT>.
+ * <P>Examples of possible stateful members are : <TT>Math.random()</TT>,
+ * <TT>System.currentTimeMillis()</TT>.
+ *
+ * <P>Stateless members of this library are always static members of the 
+ * classes, which define them. The inverse is generally not true. However,
+ * this library goes as far as assuming that static members are stateless, 
+ * if this assumption does not hold for some of Your members it is possible to 
+ * mark them as stateful using the <TT>markStateDependent()</TT> method of
+ * this class.
+ *
+ * <P>The most crucial difference between the two kind of members of this
+ * library is that evaluation of stateless methods is attempted by JEL at
+ * a compile time during the constants folding phase.
  */
 public class Library {
 
-  /**
-   * Classes, whose static methods are exported
-   */
-  private Class[] staticLib;
-  
-  /**
-   * Classes, whose virtual methods are exported
-   */
-  private Class[] dynamicLib;
+  private Hashtable names; // Hashtable(names, Hashtable(signatures,members))
   
+  private Hashtable dynIDs; // Hashtable(members, integers)
 
-  private Hashtable names;
-
-  private Hashtable dynIDs;
-
-  private Hashtable stateless;
+  private Hashtable stateless; // Hashtable(members, booleans)
   
   /**
    * Creates a library for JEL.
@@ -89,12 +83,6 @@ public class Library {
    *  methods are exported.
    */
   public Library(Class[] staticLib,Class[] dynamicLib) {
-    this.staticLib=staticLib;
-    this.dynamicLib=dynamicLib;
-    rehash();
-  };
-  
-  private void rehash() {
     names = new Hashtable();
     dynIDs = new Hashtable();
     stateless=new Hashtable();
@@ -103,75 +91,62 @@ public class Library {
     if (dynamicLib!=null)
       rehash(dynamicLib,false);
   };
-
+  
   private void rehash(Class[] arr, boolean gather_stateless) {
     for (int i=0; i<arr.length; i++) {
       Integer dynID=new Integer(i);
-      {
-	Method[] ma=arr[i].getMethods();
-	for(int j=0; j<ma.length; j++) {
-	  if (Modifier.isStatic(ma[j].getModifiers())) {
-	    if (gather_stateless)
-	      if (rehash(ma[j])) stateless.put(ma[j],Boolean.TRUE);
-	  } else {
-	    if (!gather_stateless) 
-	      if (rehash(ma[j])) dynIDs.put(ma[j],dynID); // Register ID
-	  };
-	};
-      };
 
-      {
-	Field[] fa=arr[i].getFields();
-	for(int j=0; j<fa.length; j++) {
-	  if (Modifier.isStatic(fa[j].getModifiers())) {
-	    if (gather_stateless)
-	      if (rehash(fa[j])) stateless.put(fa[j],Boolean.TRUE);
-	  } else {
-	    if (!gather_stateless)
-	      if (rehash(fa[j])) dynIDs.put(fa[j],dynID); // Register ID
-	  };
-	};
+      java.lang.reflect.Method[] marr=arr[i].getMethods();
+      java.lang.reflect.Field[] farr=arr[i].getFields();
+
+      int totalMethods=marr.length;
+      int totalMembers=totalMethods+farr.length;
+      for(int j=0;j<totalMembers;j++) {
+        Member m = j<totalMethods?
+          (Member)new Method(marr[j]):
+          (Member)new Field(farr[j-totalMethods]);
+        if ((m.getModifiers() & 0x0008)>0) { // static
+          if (gather_stateless && rehash(m)) stateless.put(m,Boolean.TRUE);
+        } else { // not static
+          if (!gather_stateless && rehash(m)) dynIDs.put(m,dynID);
+        };
       };
-      
     };
   };
 
-  private boolean rehash(Method m) {
-    return rehash(m.getName(),ExpressionImage.getSignature(m),m);
-  };
+  private boolean rehash(Member m) {
+    String name=m.getName();
+    String signature=ClassFile.getSignature(m);
 
-  private boolean rehash(Field f) {
-    // fields behave like methods with no arguments.
-    return rehash(f.getName(),
-		  "()"+ExpressionImage.getSignature(f.getType()),f);
-  };
-
-  private boolean rehash(String name, String signature, Object fm) {
+    // for the purpose of matching fields behave like methods with no
+    // arguments
+    if (m instanceof Field) signature="()"+signature;
+    
     Hashtable signatures=(Hashtable)names.get(name);
     if (signatures==null) { 
       // No method with this name was added
       Hashtable signatures_new=new Hashtable();
-      signatures_new.put(signature,fm);
+      signatures_new.put(signature,m);
       names.put(name,signatures_new);
       return true;
     };
     // Name exists in the library, check for possible signature conflict.
     Object conflicting_method=signatures.get(signature);
     if (conflicting_method==null) { // No conflict
-      signatures.put(signature,fm);
+      signatures.put(signature,m);
       return true;
     };
     if (Debug.enabled) {
       Debug.println("Conflict was detected during the library initialization."+
-		    " Conflicting "+"\""+name+signature+
-		    "\", conflicting :"+ conflicting_method+" and "+fm+" .");
+                    " Conflicting "+"\""+name+signature+
+                    "\", conflicting :"+ conflicting_method+" and "+m+" .");
     };
     // If no debug then the method is ignored.
     return false;
   };
 
   /**
-   * This function is itended for marking static function as having the internal state.
+   * This method marks a static member as having the internal state.
    * <P> If You include <TT>java.lang.Math</TT> into the library it is
    * necessary to mark <TT>java.lang.random()</TT> as having the state. This
    * can be done by calling <TT>markStateDependent("random",null)</TT>
@@ -183,7 +158,7 @@ public class Library {
    */
   public void markStateDependent(String name, Class[] params) 
        throws NoSuchMethodException {
-    Object m=getMethod(name,params);
+    Object m=getMember(name,params);
     Object removed=stateless.remove(m);
     if (Debug.enabled)
       Debug.assert(removed!=null,"State dependent methos \""+m+
@@ -196,7 +171,7 @@ public class Library {
    * @return true if the method is stateless and can be invoked at
    *    compile time.
    */
-  public boolean isStateless(Object o) {
+  public boolean isStateless(Member o) {
     return stateless.containsKey(o);
   };
   
@@ -222,7 +197,8 @@ public class Library {
    * @return the method/field object of the resolved method/field.
    * @exception NoSuchMethodException if the method can't be resolved
    */
-  Object getMethod(String name,Class[] params) throws NoSuchMethodException {
+  public Member getMember(String name,Class[] params) 
+    throws NoSuchMethodException {
     Hashtable signatures=(Hashtable) names.get(name);
 
     if (signatures==null)
@@ -231,19 +207,19 @@ public class Library {
     // Choose applicable methods
     Vector applicable_methods=new Vector();
     for(Enumeration e=signatures.elements();e.hasMoreElements();) {
-      Object cm= e.nextElement();
-      Class[] cp=getParameterTypes(cm);
+      Member cm = (Member)e.nextElement();
+      Class[] cp=cm.getParameterTypes();
       
       boolean applicable=false;
       if (params!=null) { // if no parameters requested
-	if (cp.length==params.length) {  // If number of arguments matches
-	  applicable=true;
-	  for(int i=0;((i<cp.length) && applicable);i++) {
-	    applicable=ExpressionImage.canConvertByWidening(params[i],cp[i]);
-	  };
-	};
+        if (cp.length==params.length) {  // If number of arguments matches
+          applicable=true;
+          for(int i=0;((i<cp.length) && applicable);i++) {
+            applicable=TypesStack.isWidening(params[i],cp[i]);
+          };
+        };
       } else {
-	applicable=(cp.length==0);
+        applicable=(cp.length==0);
       };
       
       if (applicable) applicable_methods.addElement(cm);
@@ -257,54 +233,40 @@ public class Library {
     };
     
     if (applicable_methods.size()==1) 
-      return applicable_methods.firstElement();
+      return (Member)applicable_methods.firstElement();
     
     // Search for the most specific method
     Enumeration e=applicable_methods.elements();
-    Object most_specific=e.nextElement();
-    Class[] most_specific_params=getParameterTypes(most_specific);
+    Member most_specific=(Member)e.nextElement();
+    Class[] most_specific_params=most_specific.getParameterTypes();
     while (e.hasMoreElements()) {
-      Object cm= e.nextElement();
-      Class[] cp=getParameterTypes(cm);
+      Member cm= (Member)e.nextElement();
+      Class[] cp=cm.getParameterTypes();
       boolean moreSpecific=true;
       boolean lessSpecific=true;
       
       for(int i=0; i<cp.length; i++) {
-	moreSpecific = moreSpecific && 
-	  ExpressionImage.canConvertByWidening(cp[i],most_specific_params[i]);
-	lessSpecific = lessSpecific &&
-	  ExpressionImage.canConvertByWidening(most_specific_params[i],cp[i]);
+        moreSpecific = moreSpecific && 
+          TypesStack.isWidening(cp[i],most_specific_params[i]);
+        lessSpecific = lessSpecific &&
+          TypesStack.isWidening(most_specific_params[i],cp[i]);
       };
       
       if (moreSpecific && (!lessSpecific)) {
-	
-	most_specific=cm;
-	most_specific_params=cp;
+        most_specific=cm;
+        most_specific_params=cp;
       };
       
       if (! (moreSpecific ^ lessSpecific))
-	throw new NoSuchMethodException("Ambiguity detected between \""+
-					describe(name,most_specific_params) +
-					"\" and \""+ describe(name,cp) +
-					"\" on invocation \""+
-					describe(name,params)+"\" .");
-      
+        throw new NoSuchMethodException("Ambiguity detected between \""+
+                                        describe(name,most_specific_params) +
+                                        "\" and \""+ describe(name,cp) +
+                                        "\" on invocation \""+
+                                        describe(name,params)+"\" .");
     };
-    
     return most_specific;
   };
 
-  private Class[] getParameterTypes(Object o) {
-    Class[] res=null;
-    if (o instanceof Field) 
-      res=new Class[0];
-    else if (o instanceof Method) 
-      res=((Method)o).getParameterTypes();
-    else if (Debug.enabled) 
-      Debug.println("Library may contain only fields or methods.");
-    return res;
-  };
-
   private String describe(String name,Class[] params) {
     String res="";
     StringBuffer invp=new StringBuffer();
@@ -330,8 +292,8 @@ public class Library {
    * @exception NullPointerException if method is not a dynamic method of
    *            this library.
    */
-  int getDynamicMethodClassID(Object fm) {
-    Integer id=(Integer)dynIDs.get(fm);
+  public int getDynamicMethodClassID(Member m) {
+    Integer id=(Integer)dynIDs.get(m);
     if (id==null) return -1;
     return id.intValue();
   };
@@ -379,8 +341,9 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Double.TYPE;
-      Method mf=(Method)ll.getMethod("round",par);
-      if ((mf!=null) && (mf.equals(math.getMethod("round",par))))
+      Member mf=ll.getMember("round",par);
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("round",par)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -391,8 +354,9 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Float.TYPE;
-      Method mf=(Method)ll.getMethod("round",par);
-      if ((mf!=null) && (mf.equals(math.getMethod("round",par))))
+      Member mf=ll.getMember("round",par);
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("round",par)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -403,13 +367,14 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Integer.TYPE;
-      Method mf=(Method)ll.getMethod("round",par);
+      Member mf=ll.getMember("round",par);
 
       Class[] par1=new Class[1];
       par1[0]=Float.TYPE;
 
 
-      if ((mf!=null) && (mf.equals(math.getMethod("round",par1))))
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("round",par1)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -420,12 +385,13 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Integer.TYPE;
-      Method mf=(Method)ll.getMethod("abs",par);
+      Member mf=ll.getMember("abs",par);
 
       Class[] par1=new Class[1];
       par1[0]=Integer.TYPE;
 
-      if ((mf!=null) && (mf.equals(math.getMethod("abs",par1))))
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("abs",par1)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -436,12 +402,13 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Byte.TYPE;
-      Method mf=(Method)ll.getMethod("abs",par);
+      Member mf=ll.getMember("abs",par);
       
       Class[] par1=new Class[1];
       par1[0]=Integer.TYPE;
       
-      if ((mf!=null) && (mf.equals(math.getMethod("abs",par1))))
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("abs",par1)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -452,12 +419,13 @@ public class Library {
     try {
       Class[] par=new Class[1];
       par[0]=Character.TYPE;
-      Method mf=(Method)ll.getMethod("abs",par);
+      Member mf=ll.getMember("abs",par);
       
       Class[] par1=new Class[1];
       par1[0]=Integer.TYPE;
       
-      if ((mf!=null) && (mf.equals(math.getMethod("abs",par1))))
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("abs",par1)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -469,13 +437,14 @@ public class Library {
       Class[] par=new Class[2];
       par[0]=Integer.TYPE;
       par[1]=Float.TYPE;
-      Method mf=(Method)ll.getMethod("min",par);
+      Member mf=ll.getMember("min",par);
       
       Class[] par1=new Class[2];
       par1[0]=Float.TYPE;
       par1[1]=Float.TYPE;
       
-      if ((mf!=null) && (mf.equals(math.getMethod("min",par1))))
+      if ((mf!=null) && 
+          (mf.equals(new Method(math.getMethod("min",par1)))))
 	  t.testOK(); else t.testFail();
     } catch (Throwable e) {
       Debug.reportThrowable(e);
@@ -485,7 +454,7 @@ public class Library {
     t.startTest("Attempt of access to the field PI");
     try {
       Class[] par=new Class[0];
-      Field f=(Field)ll.getMethod("PI",par);
+      Member f=ll.getMember("PI",par);
       if ((f!=null) && (f.getName().equals("PI")))
 	t.testOK(); else t.testFail();
     } catch (Throwable e) {
@@ -496,7 +465,7 @@ public class Library {
     t.startTest("Checking assignment of state dependence ");
     try {
       ll.markStateDependent("random",null);
-      if (!ll.isStateless(ll.getMethod("random",null)))
+      if (!ll.isStateless(ll.getMember("random",null)))
 	t.testOK();
       else
 	t.testFail();
